
'''
Cette classe a pour objectif de permettre d'augmenter le nombre de données d'un dataset de signaux EEG par
extraction de bruit et fusion avec des signaux aléatoires.

Elle permettra ainsi de pallier à la faible quantité de données qui peuvent être récoltées lors d'une session
d'acquisition de datasets.
'''

import random

import numpy as np
from scipy import signal

class EEGDataEnhancer:
    def __init__(self,data,times,labels):
        self.data = np.array(data)
        self.times = times
        self.original_size = data.shape[0]
        self.labels = labels

    # Génération des coefficients d'un filtre passe-haut de Butteworth
    def butter_highpass(self,cutoff, fs, order=5):
        nyq = 0.5 * fs
        normal_cutoff = cutoff / nyq
        b, a = signal.butter(order, normal_cutoff, btype="high", analog=False)
        return b, a

    # Filtrage des données à l'aide du filtre passe-haut
    def butter_highpass_filter(self,data, cutoff, fs, order=5):
        b, a = self.butter_highpass(cutoff, fs, order=order)
        y = signal.filtfilt(b, a, data)
        return y

    # Suppression de l'effet de bord dû au filtrage passe-haut
    def side_effect_removing(self,noise_candidate,delta):
        for i in range(0,noise_candidate.shape[0]):
            for k in range(0,delta):
                noise_mean = np.mean(noise_candidate[i, 0 + delta:noise_candidate.shape[1] - delta])
                noise_candidate[i][k] = noise_mean
                noise_candidate[i][noise_candidate.shape[1]-1 - k] = noise_mean
        return noise_candidate

    # Ajout du bruit à un signal EEG
    def add_noise(self,noise_candidate,enhance_candidate):
        delta = 22
        noise_candidate = self.side_effect_removing(np.array(noise_candidate),delta)
        enhance_candidate = np.array(enhance_candidate)
        enhance_candidate = enhance_candidate - noise_candidate
        return enhance_candidate

    # Choix d'un signal candidat à l'augmentation de manière aléatoire
    # Le numéro de l'essai sélectionnée doit être différente de exception_pos
    def choose_random_enhance_candidate(self,exception_pos):
        enhance_pos = exception_pos
        while enhance_pos == exception_pos:
            enhance_pos = random.randint(0,self.original_size-1)
        return self.data[enhance_pos], enhance_pos

    # Choix d'un signal candidat pour l'extraction du bruit de manière aléatoire
    def choose_random_noise_candidate(self):
        noise_pos = random.randint(0, self.original_size-1)
        return self.data[noise_pos],noise_pos

    # Augmentation des données de l'attribut data
    def enhance_data(self):
        for i in range(self.times-1):
            for k in range(self.original_size):
                noise_candidate,noise_pos = self.choose_random_noise_candidate()
                noise = self.butter_highpass_filter(noise_candidate, 85, 250)
                enhance_candidate, enhance_pos = self.choose_random_enhance_candidate(noise_pos)
                self.labels = np.append(self.labels,self.labels[enhance_pos])
                enhanced_data = self.add_noise(noise,enhance_candidate)
                enhanced_data = enhanced_data.reshape(1,enhanced_data.shape[0],enhanced_data.shape[1])
                self.data = np.concatenate((self.data,enhanced_data),axis=0)

    # Transformation des données de l'attribut data puis retour de celles-ci
    def get_enhanced_data(self):
        self.enhance_data()
        return self.data,self.labels





