import numpy as np
import tensorflow
from sklearn import metrics
from tensorflow.python.keras.callbacks import ModelCheckpoint
from tensorflow.python.keras.utils import np_utils

from EEGDataEnhancer import EEGDataEnhancer
from EEGModels import EEGNet, DeepConvNet, ShallowConvNet, MixedConvNet


class EEGDecodingModel:

    nb_classes = 0

    def __init__(self,model_type,chans,samples,nb_classes):
        self.nb_classes = nb_classes
        self.data_enhancer = None
        self.model_type = model_type
        self.is_trained = False
        if self.model_type == "EEGNet":
            self.model = EEGNet(nb_classes=nb_classes, Chans=chans, Samples=samples,
                           dropoutRate=0.5, kernLength=32, F1=8, D=2, F2=16,
                           dropoutType='Dropout')
        elif self.model_type == "DeepConvNet":
            self.model = DeepConvNet(nb_classes=nb_classes, Chans=chans, Samples=samples)
        elif model_type == "ShallowConvNet":
            self.model = ShallowConvNet(nb_classes=nb_classes, Chans=chans, Samples=samples)
        else:
            self.model = MixedConvNet(nb_classes=nb_classes, Chans=chans, Samples=samples,
                           dropoutRate=0.5, kernLength=32, F1=8, D=2, F2=16,
                           dropoutType='Dropout')
        print("Model " + self.model_type + " is correctly initialized.")

    def train(self,load_model,epochs,labels,epochs_number,save_as):

        print("Training model " + self.model_type + "...")
        # Extraction des données de la variable epochs
        X = epochs.get_data() * 1000  # format is in (trials, chanels, samples)
        y = labels

        self.data_enhancer = EEGDataEnhancer(X, 3,y)
        X,y = self.data_enhancer.get_enhanced_data()
        print("Data shape after augmentation : ",X.shape)

        kernels, chans, samples = 1, X.shape[1], X.shape[2]

        # Découpage des données sous la forme de jeux d'entraînement, de validation et de test
        part1 = int(X.shape[0]*0.92)
        part2 = X.shape[0]

        X_train = X[0:part1, ]
        Y_train = y[0:part1]
        X_validate = X[part1:part2, ]
        Y_validate = y[part1:part2]
        X_test = X[part1:, ]
        Y_test = y[part1:]

        Y_train = np_utils.to_categorical(Y_train - 1, self.nb_classes)
        Y_validate = np_utils.to_categorical(Y_validate - 1, self.nb_classes)
        Y_test = np_utils.to_categorical(Y_test - 1, self.nb_classes)

        # Conversion au format Essais, electrodes, données, noyaux

        X_train = X_train.reshape(X_train.shape[0], chans, samples, kernels)
        X_validate = X_validate.reshape(X_validate.shape[0], chans, samples, kernels)
        X_test = X_test.reshape(X_test.shape[0], chans, samples, kernels)

        # Définition de la fonction de loss et de l'optimiseur utilisé (par défaut adam)
        self.model.compile(loss='categorical_crossentropy', optimizer='adam',
                      metrics=['accuracy', tensorflow.keras.metrics.Precision(), tensorflow.keras.metrics.Recall()])

        # Nombre de paramètres utilisés par le modèle
        numParams = self.model.count_params()

        # Sauvegarde du meilleur modèle courant dans un fichier checkpoint.h5
        checkpointer = ModelCheckpoint(filepath=save_as, verbose=1,
                                       save_best_only=True)

        # Chargement d'un modèle préalablement sauvegardé dans un fichier h5 si demandé par l'utilisateur
        if (load_model):
            if self.model_type == "EEGNet":
                self.model.load_weights('model_eegnet.h5')
            elif self.model_type == "DeepConvNet":
                self.model.load_weights('model_deepconvnet.h5')
            else:
                self.model.load_weights('model_shallowconvnet.h5')

            # Calcul de l'accuracy du modèle chargé sur les données de test courantes
            probs = self.model.predict(X_test)
            preds = probs.argmax(axis=-1)
            acc = np.mean(preds == Y_test.argmax(axis=-1))
            print("\nClassification accuracy before training : %f " % (acc))

        class_weights = {0: 1, 1: 1, 2: 1, 3: 1}

        # Entraînement du réseau de neurones
        # Définition ici notamment de la taille de lot (batch_size) et du nombre d'epochs (epochs)
        print(X_train)
        print(Y_train)

        self.model.fit(X_train, Y_train, batch_size=1, epochs=epochs_number,
                            verbose=2, validation_data=(X_validate, Y_validate),
                            callbacks=[checkpointer], class_weight=class_weights)

        self.is_trained = True

    def load_model(self,filename):
        print("Loading " + self.model_type + " model...")
        try:
            self.model.load_weights(filename)
            print("Model succesfully loaded from file : " + filename)
        except:
            raise Exception("Unknown model weights filename.")

    def predict(self,epochs):
        X = epochs.get_data() * 1000

        probs = self.model.predict(X)
        preds = probs.argmax(axis=-1) + 1

        return preds

