import pickle

import numpy as np
from sklearn import metrics
from sklearn.svm import SVR

from EEGDataEnhancer import EEGDataEnhancer
from FBCSP_toolbox.Classifier import Classifier
from FBCSP_toolbox.FBCSP import FBCSP
from FBCSP_toolbox.MLEngine import FilterBank


class EEGDecodingModelFBCSP:
    def __init__(self,classes,sfreq):
        self.data_enhancer = None
        self.classes_classifiers = []
        self.fbcsp_train = None
        self.model_type = "FBCSP"
        self.sample_frequency = sfreq
        for i in range(classes):
            classifier_type = SVR(gamma='auto')
            classifier = Classifier(classifier_type)
            self.classes_classifiers.append(classifier)
        print("Model " + self.model_type + " is correctly initialized.")

    def train(self,epochs,labels,save_as):

        print("Training a FBCSP + SVR based model...")

        # Récupération des données d'entraînement et paramétrage de la FilterBank pour filtrer les données brutes
        data = epochs.get_data() * 1e6

        self.data_enhancer = EEGDataEnhancer(data, 3, labels)
        data, labels = self.data_enhancer.get_enhanced_data()

        print("Data shape after augmentation : ",data.shape)
        eeg_data = {'x_data': data,
                    'y_labels': labels,
                    'fs': self.sample_frequency}

        fbank = FilterBank(eeg_data.get('fs'))
        fbank_coeff = fbank.get_filter_coeff()

        # Filtrage des données EEG brutes
        X_train = fbank.filter_data(eeg_data.get('x_data'))
        y_train = eeg_data.get('y_labels')

        # Entraînement d'une FBCSP à partir des données d'entraînement
        m_filters = 2
        self.fbcsp_train = FBCSP(m_filters)
        self.fbcsp_train.fit(X_train, y_train)

        n_classes = 4
        y_classes_unique = np.unique(y_train)

        # Une valeur va être déterminée pour chacune des classes
        for j in range(n_classes):
            # Détermination de la classe qui est actuellement étudiée
            cls_of_interest = y_classes_unique[j]

            # Définition d'une fonction lambda passant à 0 les labels qui sont égaux à la classe recherchée
            # et mettant la valeur 1 dans le cas contraire
            select_class_labels = lambda cls, y_labels: [0 if y == cls else 1 for y in y_labels]

            # Conversion des labels en utilisant le lambda précédent
            y_train_cls = np.asarray(select_class_labels(cls_of_interest, y_train))

            # Extraction de caractéristiques des données d'entraînement et de test grâce à la FBCSP.
            # Attention : pour fonctionner, les données de test doivent être filtrées avec la
            # même instance FBCSP que durant l'entraînement.
            x_features_train = self.fbcsp_train.transform(X_train, class_idx=cls_of_interest - 1)

            # Entraînement du classifieur de type SVR
            self.classes_classifiers[j].fit(x_features_train, np.asarray(y_train_cls, dtype=float))

        print("Training successful !")
        print("Saving the trained model...")

        pickle.dump(self.classes_classifiers, open(save_as + ".svrlst", 'wb'))
        pickle.dump(self.fbcsp_train, open(save_as + ".fbcsp", 'wb'))

        print("Model successfully saved to files : ")
        print("- " + save_as + ".svrlst")
        print("- " + save_as + ".fbcsp")

    def predict(self,epochs,true_labels):
        test_data, test_labels = epochs, true_labels

        # Récupération des données de test
        data = test_data.get_data() * 1e6
        
        eeg_data = {'x_data': data,
                    'y_labels': test_labels,
                    'fs': self.sample_frequency}

        # Filtrage des données de test
        fbank = FilterBank(eeg_data.get('fs'))
        fbank_coeff = fbank.get_filter_coeff()
        X_test = fbank.filter_data(eeg_data.get('x_data'))
        y_test = eeg_data.get('y_labels')

        # Initialisation de tableaux
        # y_classes_unique contiendra l'ensemble des classes du dataset étudié.
        # y_test_predicted contiendront une valeur pour chaque essai et chaque classe (dans le cas du BCI
        # Competition 2008 2a, ils auront la forme d'une matrice 288x4).
        n_classes = 4
        y_classes_unique = np.unique(y_test)
        y_test_predicted = np.zeros((y_test.shape[0], n_classes), dtype=float)

        # Une valeur va être déterminée pour chacune des classes
        for j in range(n_classes):
            # Détermination de la classe qui est actuellement étudiée
            cls_of_interest = y_classes_unique[j]

            # Définition d'une fonction lambda passant à 0 les labels qui sont égaux à la classe recherchée
            # et mettant la valeur 1 dans le cas contraire
            select_class_labels = lambda cls, y_labels: [0 if y == cls else 1 for y in y_labels]

            # Conversion des labels en utilisant le lambda précédent
            y_train_cls = np.asarray(select_class_labels(cls_of_interest, y_test))

            # Extraction de caractéristiques des données d'entraînement et de test grâce à la FBCSP.
            # Attention : pour fonctionner, les données de test doivent être filtrées avec la
            # même instance FBCSP que durant l'entraînement.
            x_features_test = self.fbcsp_train.transform(X_test, class_idx=cls_of_interest - 1)

            # Réalisation d'une prédiction et collecte du résultat à l'emplacement alloué pour la classe
            # en cours d'étude dans y_test_predicted
            y_test_predicted[:, j] = self.classes_classifiers[j].predict(x_features_test)

        # La classe finale est déterminée pour chaque essai en récupérant l'indice de la classe pour laquelle
        # la valeur prédite par la fonction predict() est la plus faible.
        # La matrice y_test_predicted_multi contient alors l'ensemble des classes prédites.
        y_test_predicted_multi = np.asarray(
            [np.argmin(y_test_predicted[i, :]) for i in range(y_test_predicted.shape[0])]) + 1

        # Calcul de l'accuracy et du kappa score suite à la classification
        acc = np.sum(y_test_predicted_multi == y_test.flatten(), dtype=float) / len(y_test)
        kappa_score = metrics.cohen_kappa_score(y_test, y_test_predicted_multi)

        return y_test_predicted_multi,acc,kappa_score

    def load_model(self,filename):
        print("Loading model...")

        self.classes_classifiers = pickle.load(open(filename + ".svrlst", 'rb'))
        self.fbcsp_train = pickle.load(open(filename + ".fbcsp", 'rb'))

        print("Model successfully saved from files : ")
        print("- " + filename + ".svrlst")
        print("- " + filename + ".fbcsp")