# -*- coding: utf-8 -*-
"""
Created on Fri Oct 22 18:52:45 2021

@author: Dylan Jacquot
"""
# This program uses the BCI Competition 2008 2a dataset.

import scipy.io
import numpy as np
from os.path import join as pjoin
import mne

# Les sujets vont de A01 à A09.
# Pour obtenir les données EEG, il suffit d'indiquer un nom de sujet.
# Il existe deux versions des EEG. Si la lettre E est spécifiée à la fin du
# nom du sujet, il s'agit d'un EEG d'evaluation. Si cette lettre est un
# T il s'agit d'un EEG de train.


# Cette fonction a pour rôle de constituer un numpy array contenant les
# signaux EEG pour chaque essai d'un sujet donné ainsi que les labels
# associés à chaque essai sous la forme d'un tuple (output_data,labels)

# Définition du numpy array regroupant l'ensemble des données à 4 indices :
#    - Indice 1 : canal d'acquisition
#    - Indice 2 : numéro de l'essai 
#    - Indice 3 : choix de l'information (0 pour le temps, 1 pour la valeur du signal)
#    - Indice 4 : valeur en un point de l'information sélectionnée


# En indiquant les indices 1 à 3, on peut donc obtenir les données EEG et
# temporelles associées à un essai sur une électrode donnée
def generate_usable_data(nom_sujet):
    
    #Extraction des labels 
    
    # Constitution du chemin d'accès vers les labels
    
    mat_fname = pjoin('labels', (nom_sujet) + ".mat")
    
    # Récupération des labels sous la forme d'une liste transformée en numpy array
    # Ces labels caractérisent pour chaque session d'un sujet le geste imaginé
    
    labels = scipy.io.loadmat(mat_fname)
    labels = np.array(labels['classlabel'])
    
    # Lecture du fichier .gdf contenant les données EEG
    
    mat_fname2 = pjoin('dataset_2a', (nom_sujet + ".gdf"))
    raw = mne.io.read_raw_gdf(mat_fname2,eog=None, misc=None, preload=True)
    
    # Précisions complémentaires relatives au données : 
    
    # Canaux d'acquisition
    # 0-21 : canaux EEG
    # 22-24 : canaux EOG (utilisés pour éliminer les artéfacts)
    
    # Électrodes correspondantes dans l'ordre :
    # ['EEG-Fz', 'EEG-0', 'EEG-1', 'EEG-2', 'EEG-3', 'EEG-4', 'EEG-5', 'EEG-C3', 
    # 'EEG-6', 'EEG-Cz', 'EEG-7', 'EEG-C4', 'EEG-8', 'EEG-9', 'EEG-10', 'EEG-11', 
    # 'EEG-12', 'EEG-13', 'EEG-14', 'EEG-Pz', 'EEG-15', 'EEG-16', 'EOG-left', 
    # 'EOG-central', 'EOG-right']
    
    # Dates de début des différentes runs : 
    
    # Runs EOG :
    # Run 1 : 0s => acquisition EOG avec les yeux ouverts 
    # Run 2 : 109.04 s => acquisition EOG avec les yeux fermés
    # Run 3 : 219.372 s => acquisition EOG avec mouvement de la rétine
    
    # Runs EEG :
    # Run 4 : 368.896 s
    # Run 5 : 756.236 s
    # Run 6 : 1143.576 s
    # Run 7 : 1530.916 s
    # Run 8 : 1918.256 s
    # Run 9 : 2305.596
    
    
    # Événements importants de l'expérience : 
    # id 768 : début d'un nouvel essai
    # id 769 : mouvement imaginé de la main gauche(classe 1 dans les labels)
    # id 770 : mouvement imaginé de la main droite (classe 2 dans les labels)
    # id 771 : mouvement imaginé du pied (classe 3 dans les labels)
    # id 772 : mouvement imaginé de la langue (classe 4 dans les labels)
    # id 783 : mouvement inconnu, pour l'évaluation 
    # id 1023 : essai rejeté par les experts en raison d'un artéfact
    # id 1072 : mouveùents des yeux
    # 32766 : début d'une nouvelle run telle que définie dans la documentation
    
    # Champs essentiels des événements : 
    #   - onset : contient la valeur du début de l'événement en secondes
    #   - description : contient l'identifiant d'événement tel que décrit auparavant
    #   - duration : contient la durée de l'événement en secondes
    #   - orig_time : contient la date de début de l'événement
    
    # Récupération du vecteur de temps
    time = np.array(raw.times)
    
    # Génération des données utiles sous la forme d'une matrice numpy
    
    data_tab = []
    for channel in range(0,25):
        
        # Récupération des données EEG gloables
    
        raw_array = np.array(raw.get_data()[channel])
        
        # Constitution du tableau de données
        
        data_channel = []
        
        # Parcours de l'ensemble des événements
        for event in raw.annotations:
            
            # Sélection des événements correspondant à un début d'essai
            if event['description'] == '768':
                data_desc = []
                
                # Définition des bornes temporels de l'essai considéré
                # On utilise les champs onset et duration tels que définis
                # précédemment en commentaire.
                
                t1 = int(event['onset'])
                t2 = t1 + int(event ['duration'])
                
                # Calcul des bornes en termes de points sachant que la fréquence
                # d'échantillonnage est de 250Hz (250 echantillons par seconde)
                
                fenetre_min = t1*250
                fenetre_max = t2*250
                
                # Ajout des données temporelles puis EEG 
                data_desc.append(time[fenetre_min:fenetre_max])
                data_desc.append(raw_array[fenetre_min:fenetre_max])
                data_channel.append(data_desc)
        data_tab.append(data_channel)
           
    # Constitution du numpy array avec le tableau précédemment crée
    output_data = np.array(data_tab)
    
    return(output_data,labels)

    
