'''
Ce programme vise à calculer l'intervalle de confiance à 95% sur le temps de prédiction de chaque classifieur
de manière indépendante (EEGNet, DeepConvNet, ShallowConvNet, MixedConvNet et le classifieur SVR couplé au FBCSP) ainsi
que celui du classifieur d'ensemble.

Ces données sont importantes car c'est le temps de prédiction qui sera finalement décisif lors du passage en mode
temps-réel. De plus, bien qu'elles dépendent fortement du système sur lequel le programme est exécuté et de
l'environnement, elles vont néanmoins nous permettre de comparer les différents classifieurs entre eux et d'évaluer leur
contribution au temps de prédiction du classifieur d'ensemble.

Pour calculer ces intervalles de confiance, nous utilisons les données du dataset BCI Competition 2008 2a.
'''
import time

import numpy as np

from ComposedClassifier import ComposedClassifier
from EEGDecodingModel import EEGDecodingModel
from EEGDecodingModelFBCSP import EEGDecodingModelFBCSP
from mne_utils import get_epochs_and_labels
import scipy.stats as st

diffs = []
diffs_eegnet = []
diffs_shallow = []
diffs_deep = []
diffs_mixed = []
diffs_fbcsp = []

# Nombre de répétition d'une série de 9 prédictions par classifieur (une pour chaque sujet du dataset).
repet = 10

# Initialisation des instances d'Epochs pour chaque sujet (pour ne pas avoir à les recharger à chaque fois)
all_test_epochs_and_labels = []
for num_suj in range(1,10):
    all_test_epochs_and_labels.append(get_epochs_and_labels(num_suj, "T"))

# Répétition des séries de prédictions
for i in range(repet):

    # Réalisation d'une série de prédictions (une par sujet du dataset et par classifieur)
    for num_suj in range(1,10):
        # Chargement des données du dataset BCI Competition 2008 2a
        test_epochs, test_labels = all_test_epochs_and_labels[num_suj-1]
        test_labels = test_labels.flatten()

        # Conservation d'un fragment de 2 secondes de chaque essai du dataset
        test_epochs = test_epochs.crop(tmin=2.5,tmax=3.5)

        # Récupération du nombre d'électrodes utilisées et du nombre d'échantillons par essai
        data = np.array(test_epochs.get_data())
        chans = data.shape[1]
        samples = data.shape[2]

        # Définition du nombre de classes à déterminer et de la fréquence d'échantillonnage
        classes = 4
        sfreq = 250

        # Liste contenant l'ensemble des modèles qui vont être combinés dans le classifieur d'ensemble.
        # Il est possible de rajouter autant de classifieurs que souhaité dans cette liste.
        model_lst = [EEGDecodingModel("EEGNet", chans, samples, nb_classes=4), EEGDecodingModel("ShallowConvNet", chans, samples, nb_classes=4),
                     EEGDecodingModel("MixedConvNet", chans, samples, nb_classes=4), EEGDecodingModel("DeepConvNet", chans, samples, nb_classes=4)]

        # Définition d'un nouveau classifieur d'ensemble (pour simuler, par exemple, la fermeture de l'application)
        composed2 = ComposedClassifier(model_lst,"10221")

        # Réalisation d'une prédiction du classifieur d'ensemble et calcul du temps de prédiction
        t1 = time.time()
        _= composed2.predict(test_epochs)
        t2 = time.time()
        diff = t2 - t1

        # Ajout au tableau contenant l'ensemble des temps de prédiction du classifieur d'ensemble
        diffs.append(diff)

        # Calcul du temps de prédiction pour chaque classifieur de manière indépendante des autres
        for model in model_lst:

            # Calcul du temps de prédiction du classifieur courant
            t1 = time.time()
            _ = model.predict(test_epochs)
            t2 = time.time()
            diff = t2 - t1
            if model.model_type == "EEGNet":
                # Ajout au tableau contenant l'ensemble des temps de prédiction du classifieur EEGNet
                diffs_eegnet.append(diff)
            elif model.model_type == "ShallowConvNet":
                # Ajout au tableau contenant l'ensemble des temps de prédiction du classifieur ShallowConvNet
                diffs_shallow.append(diff)
            elif model.model_type == "DeepConvNet":
                # Ajout au tableau contenant l'ensemble des temps de prédiction du classifieur DeepConvNet
                diffs_deep.append(diff)
            elif model.model_type == "MixedConvNet":
                # Ajout au tableau contenant l'ensemble des temps de prédiction du classifieur MixedConvNet
                diffs_mixed.append(diff)

# Calcul et affichage des intervalles de confiance à 95%

print("Number of repetitions done in order to calculate the confidence intervals : ",len(diffs))
eeg_itv = st.t.interval(alpha=0.95, df=len(diffs_eegnet)-1, loc=np.mean(diffs_eegnet), scale=st.sem(diffs_eegnet))
deep_itv = st.t.interval(alpha=0.95, df=len(diffs_deep)-1, loc=np.mean(diffs_deep), scale=st.sem(diffs_deep))
shallow_itv = st.t.interval(alpha=0.95, df=len(diffs_shallow)-1, loc=np.mean(diffs_shallow), scale=st.sem(diffs_shallow))
mixed_itv = st.t.interval(alpha=0.95, df=len(diffs_mixed)-1, loc=np.mean(diffs_mixed), scale=st.sem(diffs_mixed))
global_itv = st.t.interval(alpha=0.95, df=len(diffs)-1, loc=np.mean(diffs), scale=st.sem(diffs))

print("95% confidence interval for EEGNet's prediction time : ",eeg_itv)
print("95% confidence interval for DeepConvNet's prediction time : ",deep_itv)
print("95% confidence interval for ShallowConvNet's prediction time : ",shallow_itv)
print("95% confidence interval for MixedConvNet's prediction time : ",mixed_itv)
print("95% confidence interval for the Composed Classifier's prediction time : ",global_itv)
