'''
Ce programme a pour vocation de montrer comment utiliser la classe ComposedClassifier sur le dataset BCI Competition
2008 2a pour classifier des signaux EEG de mouvements imaginés. Ce programme utilise des réseaux convolutifs ARL
(classe EEGDecodingModel) et un classifieur SVR couplé à un filtrage FBCSP (classe EEGDecodingModelFBCSP).
'''
import os
from datetime import time

import numpy as np
import tensorflow
from matplotlib import pyplot as plt
from sklearn import metrics

from ComposedClassifier import ComposedClassifier
from EEGDecodingModel import EEGDecodingModel
from EEGDecodingModelFBCSP import EEGDecodingModelFBCSP
from mne_utils import get_epochs_and_labels

print("Num GPUs Available: ", len(tensorflow.config.experimental.list_physical_devices('gpu')))

accs = []
kappas = []

fbcsp_accs = []
fbcsp_kappas = []

# Itération pour entraîner et classifier les signaux EEG sur l'ensemble des sujets du dataset
for num_suj in range(1,10):
    # Chargement des données du dataset BCI Competition 2008 2a
    epochs, labels = get_epochs_and_labels(num_suj, "E")
    test_epochs, test_labels = get_epochs_and_labels(num_suj, "T")
    labels = labels.flatten()

    # Conservation d'un fragment de 2 secondes de chaque essai du dataset
    epochs = epochs.crop(tmin=2.5,tmax=3.5)
    test_epochs = test_epochs.crop(tmin=2.5,tmax=3.5)

    # Nombre d'epoques
    epochs_number = 2

    # Pas de chargement de modèle pré-entraîné
    load = False

    # Récupération du nombre d'électrodes utilisées et du nombre d'échantillons par essai
    data = np.array(epochs.get_data())
    chans = data.shape[1]
    samples = data.shape[2]

    # Définition du nombre de classes à déterminer et de la fréquence d'échantillonnage
    classes = 4
    sfreq = 250

    # Liste contenant l'ensemble des modèles qui vont être combinés dans le classifieur d'ensemble.
    # Il est possible de rajouter autant de classifieurs que souhaité dans cette liste.
    model_lst = [EEGDecodingModel("EEGNet", chans, samples,4), EEGDecodingModel("ShallowConvNet", chans, samples,4),
                 EEGDecodingModel("MixedConvNet", chans, samples,4),EEGDecodingModel("DeepConvNet", chans, samples,4)]

    # Définition d'un nouveau classifieur d'ensemble
    composed_classifier = ComposedClassifier(model_lst,"10221")

    # Entraînement du classifieur d'ensemble
    composed_classifier.train(epochs,labels,load,epochs_number)

    # Définition d'un nouveau classifieur d'ensemble (pour simuler, par exemple, la fermeture de l'application)
    composed2 = ComposedClassifier(model_lst,"10221")

    # Chargement des modèles précédemment entraînés
    composed2.load_model()

    # Récupération de la prédiction finale sur les données d'évaluation, de l'accuracy et du score kappa
    # déterminés sur celles-ci
    final_pred = composed2.predict(test_epochs)
    final_acc = metrics.accuracy_score(test_labels, final_pred)
    final_kappa = metrics.cohen_kappa_score(test_labels, final_pred)

    print("\nAccuracy for the composed classifier : ",final_acc)
    print("Kappa score for the composed classifier : ",final_kappa)

    accs.append(final_acc)
    kappas.append(final_kappa)

# Affichages de statistiques et de graphiques pour évaluer la séquence d'entraînement réalisée

print("\nMean accuracy for FBCSP + SVR : ",np.mean(np.array(fbcsp_accs)))
print("Mean kappa score for FBCSP + SVR : ",np.mean(np.array(fbcsp_kappas)))

print("\nMean accuracy for the composed classifier : ",np.mean(np.array(accs)))
print("Mean kappa score for the composed classifier : ",np.mean(np.array(kappas)))

persons = [1,2,3,4,5,6,7,8,9]
plt.scatter(persons,accs)
plt.ylim(0,1)
plt.xlabel("Subject")
plt.ylabel("Accuracy for the composed classifier")
plt.title("Accuracy for the composed classifier")
plt.show()

plt.scatter(persons,kappas)
plt.ylim(0,1)
plt.xlabel("Subject")
plt.ylabel("Kappa score for the composed classifier")
plt.title("Kappa score for the composed classifier")
plt.show()
