
'''
Fonctions utilitaires concernant la librairie MNE et l'extraction des données du dataset BCI Competition 2008
2a.
'''
import numpy as np
import scipy.io
import os
import mne

def get_epochs_and_labels(num_suj,type_exp):
    nom_sujet = "A0" + str(num_suj) + type_exp;
    mat_fname = os.path.join('labels', (nom_sujet) + ".mat")

    # Récupération des labels sous la forme d'une liste transformée en numpy array
    # Ces labels caractérisent pour chaque session d'un sujet le geste imaginé

    labels = scipy.io.loadmat(mat_fname)
    labels = np.array(labels['classlabel'])

    # Lecture du fichier .gdf contenant les données EEG

    mat_fname2 = os.path.join('dataset_2a', (nom_sujet + ".gdf"))
    raw = mne.io.read_raw_gdf(mat_fname2, eog=None, misc=None, preload=True)

    tmin, tmax = -0.1, 7.5

    # Définition des canaux EOG (électro-oculogrammes)
    raw.set_channel_types({'EOG-left': 'eog', 'EOG-central': 'eog', 'EOG-right': 'eog'})
    raw.set_eeg_reference()

    # Définition du montage du casque pour affichage ultérieur selon la norme 1020
    montage = mne.channels.make_standard_montage(kind='standard_1020')

    # Les électrodes du fichier .gbf étant mal nommées par défaut, elles sont renommées ci-dessous selon leurs
    # noms officiels de la norme 10-20 (en se basant sur la position effective de l'électrode lors des enregistrements)
    montage.rename_channels(
        {'Fz': 'EEG-Fz', 'FC3': 'EEG-0', 'FC1': 'EEG-1', 'FCz': 'EEG-2', 'FC2': 'EEG-3', 'FC4': 'EEG-4'
            , 'C5': 'EEG-5', 'C1': 'EEG-6', 'C2': 'EEG-7', 'C6': 'EEG-8', 'CP3': 'EEG-9', 'CP1': 'EEG-10',
         'CPz': 'EEG-11', 'CP2': 'EEG-12', 'CP4': 'EEG-13', 'P1': 'EEG-14', 'P2': 'EEG-15', 'Oz': 'EEG-16',
         'C3': 'EEG-C3', 'C4': 'EEG-C4', 'Pz': 'EEG-Pz', 'Cz': 'EEG-Cz'})

    # Attribution du montage constitué au fichier raw contenant l'ensemble des signaux EEG et des événements
    raw.set_montage(montage, match_case=False)

    # Application d'un filtre passe-bande sur la plage 7-30 Hz
    raw.filter(7., 30., fir_design='firwin', skip_by_annotation='edge')

    # Extraction des événements grâce aux annotations du fichier .gdf
    events = mne.events_from_annotations(raw)

    # Définition de l'événement d'intérêt comme étant l'événement 768 (début d'un nouvel essai d'imagination motrice)
    event_id = events[1]['768']

    # Choix des canaux EEG et EOG (seulement les électrodes concernant les signaux EEG et EOG)
    # Les signaux EEG contiennent le signal utile et les signaux EOG seront utiles pour supprimer le
    # bruit d'origine oculaire (par exemple les mouvements ou clignements des yeux)
    picks = mne.pick_types(raw.info, meg=False, eeg=True, stim=False, eog=False,
                           exclude='bads')

    # Découpage des données en Epochs (intervalles de taille identique basés sur la survenance d'un événement, ici
    # l'événement 768)
    epochs = mne.Epochs(raw, events[0], event_id, tmin, tmax, picks=picks,
                        baseline=(0, 0), preload=True)

    return epochs, labels;

def import_epochs_from_file(dataset_fn):

    raw = mne.io.read_raw_gdf(dataset_fn, eog=None, misc=None, preload=True)

    tmin, tmax = -0.1, 7.5

    # Définition des canaux EOG (électro-oculogrammes)
    raw.set_channel_types({'EOG-left': 'eog', 'EOG-central': 'eog', 'EOG-right': 'eog'})
    raw.set_eeg_reference()

    # Définition du montage du casque pour affichage ultérieur selon la norme 1020
    montage = mne.channels.make_standard_montage(kind='standard_1020')

    # Les électrodes du fichier .gbf étant mal nommées par défaut, elles sont renommées ci-dessous selon leurs
    # noms officiels de la norme 10-20 (en se basant sur la position effective de l'électrode lors des enregistrements)
    montage.rename_channels(
        {'Fz': 'EEG-Fz', 'FC3': 'EEG-0', 'FC1': 'EEG-1', 'FCz': 'EEG-2', 'FC2': 'EEG-3', 'FC4': 'EEG-4'
            , 'C5': 'EEG-5', 'C1': 'EEG-6', 'C2': 'EEG-7', 'C6': 'EEG-8', 'CP3': 'EEG-9', 'CP1': 'EEG-10',
         'CPz': 'EEG-11', 'CP2': 'EEG-12', 'CP4': 'EEG-13', 'P1': 'EEG-14', 'P2': 'EEG-15', 'Oz': 'EEG-16',
         'C3': 'EEG-C3', 'C4': 'EEG-C4', 'Pz': 'EEG-Pz', 'Cz': 'EEG-Cz'})

    # Attribution du montage constitué au fichier raw contenant l'ensemble des signaux EEG et des événements
    raw.set_montage(montage, match_case=False)

    # Application d'un filtre passe-bande sur la plage 7-30 Hz
    raw.filter(7., 30., fir_design='firwin', skip_by_annotation='edge')

    # Extraction des événements grâce aux annotations du fichier .gdf
    events = mne.events_from_annotations(raw)

    # Définition de l'événement d'intérêt comme étant l'événement 768 (début d'un nouvel essai d'imagination motrice)
    event_id = events[1]['768']

    # Choix des canaux EEG et EOG (seulement les électrodes concernant les signaux EEG et EOG)
    # Les signaux EEG contiennent le signal utile et les signaux EOG seront utiles pour supprimer le
    # bruit d'origine oculaire (par exemple les mouvements ou clignements des yeux)
    picks = mne.pick_types(raw.info, meg=False, eeg=True, stim=False, eog=False,
                           exclude='bads')

    # Découpage des données en Epochs (intervalles de taille identique basés sur la survenance d'un événement, ici
    # l'événement 768)
    epochs = mne.Epochs(raw, events[0], event_id, tmin, tmax, picks=picks,
                        baseline=(0, 0), preload=True)

    return epochs

def import_raw_epochs(filename):
    raw = mne.io.read_raw_gdf(filename, eog=None, misc=None, preload=True)

    # Extraction des événements grâce aux annotations du fichier .gdf
    events = mne.events_from_annotations(raw)

    # Définition de l'événement d'intérêt comme étant l'événement 768 (début d'un nouvel essai d'imagination motrice)
    event_id = events[1]['768']

    # Choix des canaux EEG et EOG (seulement les électrodes concernant les signaux EEG et EOG)
    # Les signaux EEG contiennent le signal utile et les signaux EOG seront utiles pour supprimer le
    # bruit d'origine oculaire (par exemple les mouvements ou clignements des yeux)
    picks = mne.pick_types(raw.info, meg=False, eeg=True, stim=False, eog=False,
                           exclude='bads')

    # Découpage des données en Epochs (intervalles de taille identique basés sur la survenance d'un événement, ici
    # l'événement 768)
    epochs = mne.Epochs(raw, events[0], event_id, tmin, tmax, picks=picks,
                        baseline=(0, 0), preload=True)

    return epochs;