"""  """


""" Export to FIF :
https://www.programcreek.com/python/example/92636/mne.create_info
https://mne.tools/stable/generated/mne.create_info.html
https://mne.tools/stable/generated/mne.io.RawArray.html
https://mne.tools/stable/generated/mne.io.RawArray.html#mne.io.RawArray.export
"""

#import math
from math import floor
import threading
import time

from os import makedirs as mkdir
from os.path import exists as fileExist

import numpy as np
import serial.tools.list_ports
from pyOpenBCI import cyton
from scipy.signal import butter, lfilter

from mne.io import RawArray as mneRaw
from mne.io import read_raw_fif as readFif
from mne import create_info as mneCreateInfo
from mne import Epochs

import pickle





_api_tmp_buff = ()
_api_tmp_received = False



def echoFct(data):
    global _api_tmp_buff, _api_tmp_received
    _api_tmp_buff = data
    _api_tmp_received = True


class API:

    foldername = "eeg_data"
    extension = ".binEnsim"

    index = 0
    nb_channels = 16
    buff_len = 125
    baudRate = 115200
    function_return = False

    Freq_sampling = 125

    threadAPI = None
    board = None

    """ Initiate and connect to cyton at serial port """
    def __init__(self, serialPort, dataInterval=10):
        self.buff = np.zeros([self.nb_channels, self.buff_len])
        self.dataInterval = dataInterval
        self.board = cyton.OpenBCICyton(port=serialPort, daisy=True, baud=self.baudRate, timeout=None, max_packets_skipped=1)

    """ Initiate and connect later """
    def __init__(self, dataInterval=10):
        self.buff = np.zeros([self.nb_channels, self.buff_len])
        self.dataInterval = dataInterval

    """ Static. List all available serial ports """
    def getSerialPorts():
        lPorts = []
        ports = serial.tools.list_ports.comports()
        for port, desc, hwid in sorted(ports):
            lPorts.append([port, desc])
        return lPorts

    """ Static. Apply bandstop filter to data """
    def bandstop(self, data, lowcut, highcut, fs, order):


        nyq = 0.5 * fs
        low = lowcut / nyq
        high = highcut / nyq

        i, u = butter(order, [low, high], btype='bandstop')
        y = lfilter(i, u, data)
        return y


    
    """ Define the function called when enought data is received. Function need to be like : "def func(data):" where "data" will be a 2D shaped array """
    def setDataReceivedFunction(self, func):
        self.function_return = func

    """ Start the data stream """
    def start(self):
        if(self.function_return != False):
            print("Start request")
            self.threadAPI = threading.Thread(target=self.board.start_stream, args=(self.dataReceived,))
            self.threadAPI.start()
            
    """ Stop the data stream """
    def stop(self):
        print("stop request")
        self.board.stop_stream()
        if(self.threadAPI != None):
            self.threadAPI.join()
            self.threadAPI = None

    """ Connect to this serial port """
    def connect(self, serialPort):
        try:
            self.board = cyton.OpenBCICyton(port=serialPort, daisy=True, baud=self.baudRate, timeout=None, max_packets_skipped=1)
            return True
        except Exception:
            return False


    """ Disconnect the serial port """
    def disconnect(self):
        if(not (self.threadAPI == None)):
            self.stop()
        self.board.disconnect()

    """ Called internaly when data is received """
    def dataReceived(self, data):
        self.index += 1

        for i in range(16):
            self.buff[i][1:] = self.buff[i][:-1]
            self.buff[i][0] = data.channels_data[i]
        
        if(self.index >= self.dataInterval):
            self.buff = self.bandstop(self.buff, 45, 55, self.Freq_sampling, 1)
            self.function_return(self.buff)
            self.index = 0

    """ Resize the internal buffer to a new size. This will wipe any data stored inside the buffer when called """
    def resizeBuffer(self, newSize):
        self.buff_len = newSize
        self.buff = self.buff = np.zeros([self.nb_channels, self.buff_len])


        """ Acquire data over X seconds """
    def acquire(self, timing):
        global _api_tmp_buff, _api_tmp_received

        print("Acquire started for", timing, "seconds")
        
        tmp_size = self.buff_len
        tmp_dataInterval = self.dataInterval
        _api_tmp_received = False
        
        self.resizeBuffer(self.Freq_sampling * timing)
        self.dataInterval = self.Freq_sampling * timing

        self.setDataReceivedFunction(echoFct)

        self.start()

        self.acquireRUN = True
        
        while(_api_tmp_received != True and self.acquireRUN):
            time.sleep(0.2)

        self.stop()


        self.resizeBuffer(tmp_size)
        self.dataInterval = tmp_dataInterval

        return _api_tmp_buff

    def stopAcquire(self):
        self.acquireRUN = False



    def exportToFile(data, filename):
        if(not fileExist(API.foldername)):
            mkdir(API.foldername)

        
        with open(API.foldername+"/"+filename+API.extension, 'wb') as file:
            pickle.dump(data, file)
        
    def readFromFile(filename):
        if(not fileExist(API.foldername+"/"+filename+API.extension)):
            return None

        with open(API.foldername+"/"+filename+API.extension, 'rb') as file:
            return pickle.load(file)



    def convertDataToEpoch(data, freq=Freq_sampling):

        rawInfo = mneCreateInfo(("FP1", "FP2", "C3", "C4", "P7", "P8", "O1", "O2", "F7", "F8", "F3", "F4", "T7", "T8", "P3", "P4"), freq, ch_types="eeg") #TODO
        raw = mneRaw(data, rawInfo)
        L = len(data[0])
        epo = Epochs(raw, [[0, 1, 0]], tmin=0, tmax=(L-1)/freq, baseline=(0, (L-1)/125), on_missing="warn", reject_by_annotation=False, preload=True, verbose="WARNING")
           
        nb_epo = floor(L/freq)
        epo._data = []
           
        for i in range(nb_epo):
            epo._data.append([])
            for j in range(16):
                epo._data[i].append(data[j][i*freq:(i+1)*freq])
        epo._data = np.array(epo._data)
        return epo
        pass


""" For test purpose only """
def returnFct(data):
    print("ping")
    print(data)

""" For test purpose only """
def start(port):
    global api

    tps = 2
    api = API(dataInterval=125*tps)

    api.connect(port)
    
    api.setDataReceivedFunction(returnFct)

    api.start()

    print("go")
    
    time.sleep(20)

    print("end go")

    api.stop()

    print("end done")


""" For test purpose only """
def testFile(port):
    global api, data

    api = API()
    api.connect(port)

    data = api.acquire(8)

    api.disconnect()

    API.exportToFile(data, "test")

    return API.readFromFile("test")











