# -*- coding: utf-8 -*-
"""
Created on Sun Oct 24 16:53:02 2021

 Ce programme vise à découvrir et exploiter la librairie python mne afin de traiter plus efficacement le signal
 EEG du dataset BCI Competition 2008 et d'accéder à des fonctions utiles (par exemple affichage des signaux ou
 de graphiques/variables d'intérêt pour analyser les données.

@author: dylan
"""
import scipy.io
from os.path import join as pjoin
import mne
import numpy as np
from matplotlib import pyplot as plt
from mne import SourceEstimate
from mne.decoding import CSP
from mne.viz import plot_epochs_psd_topomap
from sklearn import metrics
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis
from sklearn.model_selection import ShuffleSplit, cross_val_score
from sklearn.pipeline import Pipeline

nom_sujet = "A01E";
mat_fname = pjoin('labels', (nom_sujet) + ".mat")

# Récupération des labels sous la forme d'une liste transformée en numpy array
# Ces labels caractérisent pour chaque session d'un sujet le geste imaginé

labels = scipy.io.loadmat(mat_fname)
labels = np.array(labels['classlabel'])

# Lecture du fichier .gdf contenant les données EEG

mat_fname2 = pjoin('dataset_2a', (nom_sujet + ".gdf"))
raw = mne.io.read_raw_gdf(mat_fname2, eog=None, misc=None, preload=True)

tmin, tmax = -0.1, 7.5

# Définition des canaux EOG (électro-oculogrammes)
raw.set_channel_types({'EOG-left': 'eog','EOG-central': 'eog','EOG-right': 'eog'})
raw.set_eeg_reference()

# Définition du montage du casque pour affichage ultérieur selon la norme 1020
montage = mne.channels.make_standard_montage(kind='standard_1020')

# Les électrodes du fichier .gbf étant mal nommées par défaut, elles sont renommées ci-dessous selon leurs
# noms officiels de la norme 10-20 (en se basant sur la position effective de l'électrode lors des enregistrements)
montage.rename_channels({'Fz':'EEG-Fz','FC3':'EEG-0','FC1':'EEG-1','FCz':'EEG-2','FC2':'EEG-3','FC4':'EEG-4'
                            ,'C5':'EEG-5','C1':'EEG-6','C2':'EEG-7','C6':'EEG-8','CP3':'EEG-9','CP1':'EEG-10',
                         'CPz':'EEG-11','CP2':'EEG-12','CP4':'EEG-13','P1':'EEG-14','P2':'EEG-15','Oz':'EEG-16',
                         'C3':'EEG-C3','C4':'EEG-C4','Pz':'EEG-Pz','Cz':'EEG-Cz'})

# Attribution du montage constitué au fichier raw contenant l'ensemble des signaux EEG et des événements
raw.set_montage(montage,match_case = False)

# Application d'un filtre passe-bande sur la plage 7-30 Hz
raw.filter(7.,30.,fir_design='firwin',skip_by_annotation='edge')

# Extraction des événements grâce aux annotations du fichier .gdf
events = mne.events_from_annotations(raw)

# Définition de l'événement d'intérêt comme étant l'événement 768 (début d'un nouvel essai d'imagination motrice)
event_id = events[1]['768']

# Choix des canaux EEG et EOG (seulement les électrodes concernant les signaux EEG et EOG)
# Les signaux EEG contiennent le signal utile et les signaux EOG seront utiles pour supprimer le
# bruit d'origine oculaire (par exemple les mouvements ou clignements des yeux)
picks = mne.pick_types(raw.info, meg=False, eeg=True, stim=False, eog=True,
                       exclude='bads')

# Découpage des données en Epochs (intervalles de taille identique basés sur la survenance d'un événement, ici
# l'événement 768)
epochs = mne.Epochs(raw, events[0], event_id, tmin, tmax, picks=picks,
                    baseline=(0, 0), preload=True)

# Définition d'un sous-ensemble d'entraînement pour la classification
epochs_train = epochs

labels = labels.flatten()

scores = []

# Récupération des données globales et d'entraînement
epochs_data = epochs.get_data()
epochs_data_train = epochs_train.get_data()
epochs_data_train_sv = epochs_train.get_data()

# Mise en place de la cross-validation pour l'évaluation des performances de classification
cv = ShuffleSplit(10, test_size=0.2, random_state=10)
cv_split = cv.split(epochs_data_train)

# Instanciation d'un classifieur LDA (Linear Discriminant Analysis) et d'un filtre adaptatif CSP (Common Spatial
# Patterns)
lda = LinearDiscriminantAnalysis()
csp = CSP(n_components=4, reg=None, log=True, norm_trace=False)


# Mise en place du classifieur avec le filtre CSP
clf = Pipeline([('CSP', csp), ('LDA', lda)])
for i in range(0,2):
    clf.fit(epochs_data,labels)

probs = clf.predict(epochs.get_data())
acc = metrics.accuracy_score(labels.flatten(), probs.flatten())

# Affichage des résultats après classfication
print("\n\n\nClassification accuracy: %f \n\n\n" %(acc))


# Affichage des patternes CSP de manière graphique
csp.fit_transform(epochs_data, labels)

# Calcul des evoked : moyennage temporel des epochs
evoked = epochs.average()

# Suppression du bruit phisiologique grâce au SSP
evoked.add_proj(mne.compute_proj_evoked(evoked.copy().crop(tmax=0)))
evoked.apply_proj()

mne.preprocessing.fix_stim_artifact(evoked)

# Correction des délais dûs au matériel utilisé
evoked.shift_time(-0.004)


times = np.arange(0, 7.5, 0.01)
fig, anim = evoked.animate_topomap(
    times=times, ch_type='eeg', frame_rate=200, time_unit='s', blit=False)
plt.show()
anim.save("topomap_animation.gif")

plot_epochs_psd_topomap(epochs)

